import config
import util
import cluster

import re
import os
import sys
from pprint import pprint
from string import Template

import boto
from boto.ec2.autoscale import AutoScaleConnection
from boto.ec2.autoscale import LaunchConfiguration

_cfg = config.Config()

class LCTemplate(Template):
    delimiter = '@'

# cloud-init shell script template
cloud_init_script = '''
#!/bin/bash

# Initialize a base system, provision via RPM installation
# Read by LaunchConfig

CLUSTER_NAME="@cluster_name"
ROLE_NAME="@role_name"

# Our stuff goes in here
BOOTSTRAP_DIR=/root/.udo
mkdir -p $BOOTSTRAP_DIR

# set up Yum S3 IAM plugin
if [[ -n "@yum_plugin_url" ]]; then
    curl @yum_plugin_url \
        > $BOOTSTRAP_DIR/yum-plugin-s3-iam.noarch.rpm
    rpm -i $BOOTSTRAP_DIR/yum-plugin-s3-iam.noarch.rpm
fi

# add our app yum repo
if [[ -n "@repo_url" ]]; then
    cat > /etc/yum.repos.d/@{app_name}.repo <<YUMREPO
[@app_name]
name=@app_name
baseurl=@repo_url
enabled=1
s3_enabled=1
gpgcheck=0
YUMREPO
fi

# load up repo metadata
yum makecache

# your stuff from cloud_init config
# pre-package-setup
@cloud_init_pre

######

# install updates
yum update -y

## install packages
# base system
if [[ -n "@base_packages" ]]; then
    yum install -y @base_packages
fi
# role packages
if [[ -n "@role_packages" ]]; then
    yum install -y @role_packages
fi

# custom post-init hook
@cloud_init_post


# might not be a bad idea to reboot after updating everything?
#/sbin/reboot now
'''

#####


class LaunchConfig:
    def __init__(self, cluster_name, role_name):
        self.cluster_name = cluster_name
        self.role_name = role_name
        self.role_config = config.get_role_config(cluster_name, role_name)
        self.conn = util.as_conn()
        self._name = "-".join([self.cluster_name, self.role_name])

    # have a default name, but can be overridden
    def name(self):
        return self._name
    def set_name(self, name):
        self._name = name

    # processes the cloud-init template, returns a string
    def cloud_init_script(self):
        # load cloud-init script template
        cloud_init_template = LCTemplate(cloud_init_script)

        cloud_init_config = _cfg.get_root()

        # add extra template vars
        cloud_init_config['base_packages'] = " ".join(_cfg.get('packages')) or ''
        cloud_init_config['yum_plugin_url'] = _cfg.get('repo', 'plugin_url') or ''
        # from role config
        cloud_init_config['role_packages'] = " ".join(self.role_config.get('packages')) or ''
        cloud_init_config['repo_url'] = self.role_config.get('repo_url') or ''

        # append extra commands from config
        cloud_init_config['cloud_init_pre'] = _cfg.get('cloud_init') or _cfg.get('cloud_init_pre') or ''
        cloud_init_config['cloud_init_post'] = _cfg.get('cloud_init_post') or ''

        cloud_init_config['cluster_name'] = self.cluster_name or ''
        cloud_init_config['role_name'] = self.role_name or ''

        cloud_init = cloud_init_template.substitute(**cloud_init_config)
        return cloud_init

    # does a LC exist with our name?
    def exists(self):
        lc = self.get_lc()
        if lc:
            return True
        return False

    # we can't modify a launchconfig in place, we have to create
    # a new one. returns new udo.lc
    def update(self):
        if not self.exists():
            # easy, just create it
            print "not exists"
            self.activate()
            return self
        # generate a name for the new lc version
        name = self.name()
        vermatch = re.search(r'-v(\d+)$', name)
        if vermatch:
            # increment version #
            ver = int(vermatch.group(1))
            ver = ver+1
            name = re.sub(r'-v(\d+)$', '-v'+str(ver), name)
        else:
            name = name + '-v2'
        # create the new lc and return it
        newlc = LaunchConfig(self.cluster_name, self.role_name)
        newlc.set_name(name)
        newlc.activate()
        return newlc

    def get_lc(self):
        conn = util.as_conn()
        lcs = conn.get_all_launch_configurations(names = [self.name()])
        if not len(lcs):
            return None
        return lcs[0]

    def deactivate(self):
        if not self.exists():
            return
        print "Deleting launchconfig..."
        lc = self.get_lc()
        if util.retry(lambda: lc.delete(), 500):
            util.message_integrations("Deleted LaunchConfig {}".format(self.name()))
        else:
            util.message_integrations("Failed to delete LaunchConfig {}".format(self.name()))

    # creates the LaunchConfig
    # returns True if LC exists
    def activate(self):
        conn = util.as_conn()
        conn = boto.ec2.autoscale.connect_to_region('us-west-2')

        name = self.name()

        # check if this LC already exists
        if self.exists():
            if not util.confirm("LaunchConfig {} already exists, overwrite?".format(name)):
                return True
            # delete existing
            conn.delete_launch_configuration(name)

        # get configuration for this LC
        cfg = self.role_config
        lc = LaunchConfiguration(
            name = name,
            image_id = cfg.get('ami'),
            instance_profile_name = cfg.get('iam_profile'),
            instance_type = cfg.get('instance_type'),
            security_groups = cfg.get('security_groups'),
            key_name = cfg.get('keypair_name'),
            user_data = self.cloud_init_script(),
            associate_public_ip_address = True,  # this is required for your shit to actually work
        )
        if not conn.create_launch_configuration(lc):
            print "Error creating LaunchConfig {}".format(name)
            return False

        util.message_integrations("Activated LaunchConfig {}".format(name))

        return lc



import yaml
from pprint import pprint
import os
import sys

"""
This class provides an interface into the entire AWS configuration.
It does more than simply convert a config file into a python data structure;
when you request a key, it will merge all values it finds downwards from the
root, using values defined at the root as defaults.
In addition, hashes and arrays are combined as it merges down.

If this is unclear, take a look at test_config.py and config.sample.yml. 

how2use:
    cfg = Config()

    # get a root configuration value
    region = cfg.get('region')

    # get a role config object
    role_cfg = cfg.new_root('clusters', 'prod', 'roles', 'webapp')

    # really queries clusters.prod.roles.webapp.packages, merging any packages 
    # array values it finds along the way
    role_cfg.get('packages')  
""" 

# class method
def load():
    file_path = get_config_file_path()
    try:
        f = open(file_path, 'r')
    except IOError as e:
        print "Failed opening file {}: {}".format(file_path, os.strerror(e.errno))
        return None
    contents = f.read()
    f.close()
    return parse(contents)

def get_config_file_path():
    # look for config file in current working directory
    cwd = os.getcwd()
    path = cwd + "/" + "udo.yml"
    return path    

# given a YAML string, parse it into a python data structure
def parse(config_yaml):
    return yaml.load(config_yaml)

class Config:
    _root = None
    _path = None

    def __init__(self, root=None):
        if root:
            self._root = root
        else:
            self._root = load()

        if not self._root:
            # loading config failed, can't proceed
            print "No udo.yml configuration file was found."
            print "Please create one or change to the directory containing your configuration."
            sys.exit(1)

    def clone(self):
        return Config(self._root)

    # returns a new Config with its base path set to *base
    def new_root(self, *base):
        new_cfg = self.clone()
        if self._path:
            base = self._path + base
        new_cfg._path = base
        return new_cfg

    # get value at current path root
    def get_root(self):
        return self.get()

    def get(self, *path):
        if self._path:
            path = self._path + path
            # print "Current path is: {}".format(path)

        cfg = self._root

        # no path specified? return root
        if len(path) == 0:
            return cfg

        # if path is only one segment, just return whatever is there
        # if len(path) == 1:
        #     return cfg.get(path[0])

        # perform merge from root to leaf node
        key = path[-1]  # get last element, this is what we're merging at all levels

        # iterate downwards, merging all values we find
        cur_val = None

        # start at root level
        cur_level = cfg

        cur_level_key = 'root'
        for level in path:
            if not cur_level:
                # dead end traversing downwards. go with whatever we found last
                print "Warning: {} is not defined in configuration at {}".format(key, level)
                return cur_val

            if key in cur_level:
                cur_val = self.merge(cur_val, cur_level.get(key))
                # print "Found {} in level {}".format(key, cur_level_key)
            
            # go one level deeper
            cur_level = cur_level.get(level)
            cur_level_key = level

        return cur_val

    # returns combination of val1+val2:
    # if val1 or val2 is None, returns whichever is not None
    # if val1 and val2 are lists, returns a combined list
    # if val1 and val2 are dict, returns a combined dict
    # if val1 and val2 are different non-None types, explodes
    # should this be recurse? not sure
    def merge(self, val1, val2):
        if val1 is None:
            return val2
        if val2 is None:
            return val1

        if val1.__class__ is not val2.__class__:
            print "{} is not the same type as {}".format(val1.__class__, val2.__class__)
            return val1

        if val1.__class__ is list:
            # merge lists
            return (val1 + val2)

        if val1.__class__ is dict:
            # merge dicts
            merged = dict(val1.items() + val2.items())
            # print "merged: {}".format(merged)
            return merged

        # assume scalar, val2 overwrites val1
        return val2

    def get_cluster_config(self, cluster_name):
        # check clusters
        cluster_config = self.new_root('clusters')
        if cluster_name not in cluster_config.get():
            print "Invalid cluster name: {}".format(cluster_name)
            return
        return cluster_config.new_root(cluster_name)

    def get_role_config(self, cluster_name, role_name):
        # check clusters
        cluster_config = self.new_root('clusters')
        if cluster_name not in cluster_config.get():
            print "Invalid cluster name: {}".format(cluster_name)
            return

        # check roles
        roles_config = cluster_config.new_root(cluster_name, 'roles')
        if not roles_config:
            print "No roles defined in cluster {}".format(cluster_name)
            return
        if role_name not in roles_config.get():
            print "Invalid role name: {}".format(role_name)
            return;

        return roles_config.new_root(role_name)



##

# for convenience        
def get_cluster_config(cluster_name):
    cluster_config = Config()
    return cluster_config.get_cluster_config(cluster_name)

def get_role_config(cluster_name, role_name):
    cluster_config = Config()
    return cluster_config.get_role_config(cluster_name, role_name)




#!/usr/bin/env python

import sys
import warnings
import os
import argparse
from pprint import pprint

import cluster
import launchconfig
import util
import asgroup
import config
import deploy

#####


# top-level commands go here
class Udo:
    def cluster(self, *args):
        args = list(args)
        if not len(args) or not args[0]:
            print "cluster command requires an action. Valid actions are: "
            print " list\n status"
            return
        action = args.pop(0)

        if action == 'list':
            cluster.list()
        else:
            # actions that require a cluster name
            if not len(args) or not args[0]:
                print "cluster name required for {}".format(action)
                return
            cluster_name = args.pop(0)
            cl = cluster.Cluster(cluster_name)
            if action == 'status':
                print "{} status: {}".format(cluster_name, cl.status())
            elif action == 'create':
                if not cl.create():
                    print "Failed to bring up {} cluster".format(cluster_name)
            else:
                print "Unknown cluster command: {}".format(action)


    # launchconfig
    def lc(self, *args):
        args = list(args)
        if not len(args) or not args[0]:
            print "launchconfig command requires an action. Valid actions are: "
            print " cloudinit (cluster) (role) - view cloud_init bootstrap script"
            print " create (cluster) (role) - create launch configuration"
            print " destroy (cluster) (role) - delete launch configuration"
            return
        action = args.pop(0)

        cluster,role = self.get_cluster_and_role_from_args(*args)
        if not cluster or not role:
            return

        lc = launchconfig.LaunchConfig(cluster, role)

        if action == 'cloudinit':
            cloudinit = lc.cloud_init_script()
            print cloudinit
        elif action == 'create':
            lc.activate()
        elif action == 'destroy':
            lc.deactivate()
        else:
            print "Unrecognized launchconfig action"


    # autoscale
    def asg(self, *args):
        args = list(args)
        if not len(args) or not args[0]:
            print "asgroup command requires an action. Valid actions are: "
            print " create (cluster) (role) - create an autoscale group"
            print " destroy (cluster) (role) - delete an autoscale group and terminate all instances"
            print " reload (cluster) (role) - destroys asgroup and launchconfig, then recreates them"
            print " updatelc (cluster) (role) - generates a new launchconfig version"
            print " scale (cluster) (role) - view current scaling settings"
            print " scale (cluster) (role) (desired) - set desired number of instances"
            return
        action = args.pop(0)

        # TODO: hook up 'list'

        cluster,role,extra = self.get_cluster_and_role_from_args(*args)
        if not cluster or not role:
            return

        ag = asgroup.AutoscaleGroup(cluster, role)

        if action == 'create':
            ag.activate()
        elif action == 'destroy':
            ag.deactivate()
        elif action == 'reload':
            ag.reload()
        elif action == 'updatelc':
            ag.update_lc()
        elif action == 'scale':
            # get scale arg
            if extra:
                scale = int(extra)
                ag.scale(scale)
            else:
                ag.get_scale_size()
        else:
            print "Unrecognized asgroup action {}".format(action)


    # CodeDeploy
    def deploy(self, *args):
        args = list(args)
        if not len(args) or not args[0]:
            print "deploy command requires an action. Valid actions are: "
            print " list applications"
            print " list groups [application]"
            print " list deployments"
            print " list configs"
            print " create (group) (commit_id)"
            print " last [group]"
            return
        action = args.pop(0)

        if action == 'list':
            dep = deploy.Deploy()
            if not len(args):
                print "list what? applications, groups, deployments or configs?"
                return
            what = args.pop(0)
            if what == 'applications' or what == 'apps':
                dep.list_applications()
            elif what == 'groups':
                # application name?
                application = None
                if len(args):
                    application = args.pop(0)
                dep.list_groups(application)
            elif what == 'configs':
                dep.list_configs()
            elif what == 'deployments':
                dep.list_deployments()
            else:
                print "Unknown list type: {}".format(what)
        elif action == 'create':
            # require group, commit_id
            if len(args) != 2:
                print "deploy create requires group and commit id"
                return

            group = args.pop(0)
            commit_id = args.pop(0)
            dep = deploy.Deploy()
            dep.create(group, commit_id)
        elif action == 'last':
            dep = deploy.Deploy()
            group = None
            if len(args) == 1:
                group = args.pop(0)
                dep.print_last_deployment(deployment_group_name=group)
            else:
                dep.print_last_deployment()
        elif action == 'stop':
            dep = deploy.Deploy()
            dep.stop_deployment()
        elif len(args) == 1:
            # let's just assume we wanna create a deployment
            group = action
            commit_id = args.pop(0)
            dep = deploy.Deploy()
            dep.create(group, commit_id)
        else:
            print "Unknown deploy command: {}".format(action)


    # for testing features
    def test(self, *args):
        args = list(args)
        if not len(args) or not args[0]:
            print "test command requires an action. Valid actions are: "
            print " integrations"
            return
        action = args.pop(0)

        if action == 'integrations':
            util.message_integrations("Testing Udo integrations")
        else:
            print "Unknown test command: {}".format(action)

    def get_cluster_and_role_from_args(self, *args):
        args = list(args)

        # need cluster/role
        if len(args) < 1:
            print "Please specify cluster name for this command"
            return None,None
        cluster = args.pop(0)

        # use role name if specified, otherwise assume they meant the obvious thing
        # if there's only one role
        if len(args):
            role = args.pop(0)
        else:
            roles = config.get_cluster_config(cluster).get('roles')
            if not roles:
                print "Cluster config for {} not found".format(cluster)
                return None,None

            rolenames = roles.keys()
            if len(rolenames) == 1:
                # assume the only role
                print "No role specified, assuming {}".format(rolenames[0])
                role = rolenames[0]
            else:
                print "Multiple roles available for cluster {}".format(cluster)
                for r in rolenames:
                    print "  - {}".format(r)
                return None,None

        # still stuff?
        extra = None
        if len(args):
            extra = args.pop(0)

        return cluster, role, extra
#####


def invoke_console():
    # argument parsing
    parser = argparse.ArgumentParser(description='Manage AWS clusters.')
    parser.add_argument('cmd', metavar='command', type=str, nargs='?',
                       help='Action to perform. Valid actions: status.')
    parser.add_argument('cmd_args', metavar='args', type=str, nargs='*',
                       help='Additional arguments for command.')
    args = parser.parse_args()
    
    if args.cmd not in dir(Udo):
        if args.cmd:
            print "'{}' is not a valid command".format(args.cmd)
        else:
            print "You must specify a command"
        # full command summary
        print """
Valid commands are:
  * cluster list - view state of clusters
  * cluster status - view state of a cluster
  * cluster create - create a VPC
  * lc cloudinit - display cloud-init script
  * lc create - create a launch configuration
  * lc destroy - delete a launch configuration
  * asg reload - destroy and create an autoscaling group to update the config
  * asg create - create an autoscaling group
  * asg destroy - delete an autoscaling group
  * asg updatelc - updates launchconfiguration in-place
  * asg scale - set desired number of instances
  * deploy list apps - view CodeDeploy applications
  * deploy list groups - view CodeDeploy application deployment groups
  * deploy list deployments - view CodeDeploy deployment statuses
  * deploy list configs - view CodeDeploy configurations
  * deploy create (group) (commit) - create new deployment for commit on group
  * deploy last - shows status of most recent deployment
  * deploy stop - cancel last deployment
        """
        sys.exit(1)

    # execute cmd
    exe = Udo()
    method = getattr(exe, args.cmd)
    method(*args.cmd_args)



if __name__ == '__main__':
    invoke_console()
